from django.db import models
from debits.debits_base.base import Period
from debits.debits_base.models import Product, SubscriptionPurchase


class PricingPlan(models.Model):
    """Pricing plan (like "Item 1", $10/month)."""
    product = models.ForeignKey(Product, on_delete=models.CASCADE)
    """Sold product."""

    name = models.CharField(max_length=255)
    """Pricing plan name."""

    price = models.DecimalField(max_digits=10, decimal_places=2)
    """The price of each recurring payment."""

    currency = models.CharField(max_length=3)
    """The currency of payments."""

    period = Period()
    """Recurring payments period."""

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<PricingPlan: %s, %s>" % ((("pk=%d" % self.pk) if self.pk else "no pk"), self.__str__())


class MyPurchase(SubscriptionPurchase):
    """An example purchase."""

    plan = models.ForeignKey(PricingPlan, on_delete=models.CASCADE)
    """The pricing plan for the purchase."""

    for_organization = models.ForeignKey('Organization', null=True, related_name='for_purchase', on_delete=models.CASCADE)
    """The organization for which the purchase was initiated.
    
    Don't mess :attr:`for_organization` with :attr:`organization`!"""

    def __repr__(self):
        return "<MyPurchase: %s>" % (("pk=%d" % self.pk) if self.pk else "no pk")


class Organization(models.Model):
    """An example organization."""

    name = models.CharField(max_length=255)
    """Organization name."""

    purchase = models.OneToOneField(MyPurchase, on_delete=models.CASCADE)
    """The current active :class:`~debits.debits_test.models.MyPurchase` for the organization."""

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Organization: %s, %s>" % ((("pk=%d" % self.pk) if self.pk else "no pk"), self.__str__())



import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

SECRET_KEY = 'Eech4Ak6Iedah1ahahMaeng4mahsee7Z'

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'debits.debits_base',
    'debits.paypal',
    'debits.debits_test',
]

MIDDLEWARE = [
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'obican', 'templates'),
                os.path.join(BASE_DIR, 'core', 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# Database
# https://docs.djangoproject.com/en/1.10/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

ROOT_URLCONF = 'debits.debits_test.urls'

# Internationalization
# https://docs.djangoproject.com/en/1.10/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

## Payments settings

#PAYMENTS_HOST = 'XXX'
#IPN_HOST = 'XXX'
#FROM_EMAIL = 'XXX'
PROLONG_PAYMENT_VIEW = 'transaction-prolong-payment'
PAYMENTS_DAYS_BEFORE_DUE_REMIND = 10
PAYMENTS_DAYS_BEFORE_TRIAL_END_REMIND = 10
#PAYPAL_EMAIL = 'XXX'
#PAYPAL_ID = 'XXX' #PayPal account ID
# https://developer.paypal.com/developer/applications
#PAYPAL_CLIENT_ID = 'XXX'
#PAYPAL_SECRET = 'XXX'
PAYPAL_DEBUG = True
PAYMENTS_REALM = 'testapp1'

try:
    from local_settings import *
except ImportError as e:
    pass



import debits.debits_base
from debits.debits_base.models import SubscriptionItem
import abc
import datetime
from django.http import HttpResponse
try:
    from html import escape  # python 3.x
except ImportError:
    from cgi import escape  # python 2.x
import debits.debits_base


def hidden_field(f, v):
    """Internal."""
    return "<input type='hidden' name='%s' value='%s'/>" % (escape(f), escape(v))


class BasePaymentProcessor(abc.ABC):
    """Executing a transaction for a particular payment processor (by a derived class).

    We receive a derivative of :class:`~debits.debits_base.models.BaseTransaction` object
    and a hash (see for example PayPay documentation) from user.

    Then the hash is amended (for example added the price from the transaction object) and
    passed to the payment processor.
    """
    @abc.abstractmethod
    def amend_hash_new_purchase(self, transaction, hash):
        """Internal."""
        pass

    def amend_hash_change_subscription(self, transaction, hash):
        """Internal."""
        raise NotImplementedError()

    def change_subscription(self, transaction, hash):
        """Start the process of changing a subscription with given hash and transaction."""
        hash = self.amend_hash_change_subscription(transaction, hash)
        return self.redirect_to_processor(hash)

    def make_purchase(self, hash, transaction):
        """Start the process of purchase with given hash and transaction."""
        hash = self.amend_hash_new_purchase(transaction, hash)
        return self.redirect_to_processor(hash)

    def make_purchase_from_form(self, hash, transaction):
        """Start the process of purchase with hash received from a HTML form and transaction."""
        hash = dict(hash)
        del hash['csrfmiddlewaretoken']
        # immediately before redirect to the processor
        return self.make_purchase(hash, transaction)

    def change_subscription_from_form(self, hash):
        """Start the process of changing a subscription with hash received from a HTML form and transaction."""
        hash = dict(hash)
        transaction = debits.debits_base.models.Item.objects.get(hash['arcamens_purchaseid'])
        del hash['arcamens_purchaseid']
        hash = self.amend_hash_change_subscription(transaction, hash)
        return self.change_subscription(transaction, hash)

    def redirect_to_processor(self, hash):
        """Internal."""
        return HttpResponse(BasePaymentProcessor.html(hash))

    # Internal
    # Use this instead of a redirect because we prefer POST over GET
    @staticmethod
    def html(hash):
        """Internal."""
        action = escape(hash['arcamens_action'])
        del hash['arcamens_action']
        return "<html><head><meta charset='utf-8'' /></head>\n" +\
            "<body onload='document.forms[0].submit()'>\n<p>Redirecting...</p>\n" + \
            "<form method='post' action='"+action+"'>\n" + \
            '\n'.join([hidden_field(i[0], str(i[1])) for i in hash.items()]) + \
            "\n</form></body></html>"

    def ready_for_subscription(self, transaction):
        """Check if ready for subscription.

        If we are in manual recurring mode, we can be not ready for subscription,
        because some payment processors (PayPal) don't allow to delay the first
        payment of a subscription for more than :meth:`self.subscription_allowed_date` days."""
        return datetime.date.today() >= self.subscription_allowed_date(transaction)

    @abc.abstractmethod
    def subscription_allowed_date(self, transaction):
        """See :meth:`ready_for_subscription`."""
        pass

    def product_name(self, purchase):
        """Internal."""
        return purchase.item.product.name


PAYMENT_PROCESSOR_AVANGATE = 1
PAYMENT_PROCESSOR_PAYPAL = 2
PAYMENT_PROCESSOR_BRAINTREE = 3
PAYMENT_PROCESSOR_DALPAY = 4
PAYMENT_PROCESSOR_RECURLY = 5


class PaymentCallback(object):
    """Mixin this class to make callbacks on a payment.

    In current implementation, :meth:`on_subscription_created` may be called when it was already started
    and :meth:`on_subscription_canceled` may be called when it is already stopped.
    (In other words, they can be called multiple times in a row.)
    """
    def on_payment(self, payment):
        """Called on any payment (subscription or regular)."""
        pass

    # def on_upgrade_subscription(self, transaction, old_subscription):
    #     pass

    def on_subscription_created(self, POST, subscription):
        """Called when a subscription is created."""
        pass

    def on_subscription_canceled(self, POST, subscription):
        """Called when a subscription is canceled."""
        pass



import abc
import hmac
import datetime

import html2text
from django.apps import apps
from django.urls import reverse
from django.db import models
from django.db.models import F
import django.db
from django.db import transaction
from django.core.exceptions import ObjectDoesNotExist
from django.core.mail import send_mail
from django.template.loader import render_to_string
from django.utils.translation import ugettext_lazy as _
from composite_field import CompositeField
from django.conf import settings

from debits.debits_base.base import logger, Period, period_to_delta


class ModelRef(CompositeField):
    """Reference to a Django model"""

    app_label = models.CharField(_('Django app with the model'), max_length=100)
    """Django app with the model."""

    model = models.CharField(_('Python model class name'), max_length=100)
    """The model class name."""


# The following function does not work as a method, because
# CompositeField is replaced with composite_field.base.CompositeField.Proxy:

def model_from_ref(model_ref):
    """Retrieves a model from `ModelRef`.

    Args:
        model_ref: A `ModelRef` field instance.

    Returns:
        A Django model class.
    """
    return apps.get_model(model_ref.app_label, model_ref.model)


class PaymentProcessor(models.Model):
    """Payment processor (such as PayPal, DalPay, etc.)"""

    name = models.CharField(_('The name of the company'), max_length=255)
    """The name of the payment processing company or service."""

    url = models.URLField(max_length=255)
    """The site of the payment processor."""

    klass = ModelRef()
    """The Django model which handles API for payments and similar stuff."""

    def __str__(self):
        return self.name


class Product(models.Model):
    name = models.CharField(_('Product name'), max_length=255)
    """Product name."""

    def __str__(self):
        return self.name


class BaseTransaction(models.Model):
    """A redirect (or other query) to the payment processor.

    It may be paid or (not yet) paid."""

    # class Meta:
    #     abstract = True

    processor = models.ForeignKey(PaymentProcessor, on_delete=models.CASCADE)
    """Payment processor."""

    creation_date = models.DateTimeField(auto_now_add=True)
    """Date of the redirect."""

    purchase = models.ForeignKey('Purchase', related_name='transactions', null=False, on_delete=models.CASCADE)
    """The stuff sold by this transaction."""

    def __repr__(self):
        return "<BaseTransaction: %s>" % (("pk=%d" % self.pk) if self.pk else "no pk")

    @staticmethod
    def custom_from_pk(pk):
        """Secret code of a transaction.

        Secret can be known only to one who created a BaseTransaction.
        This prevents third parties to make fake IPNs from a payment processor.

        Args:
            pk: the serial primary key (of :class:`BaseTransaction`) used to calculate the secret transaction code.

        Returns:
            A secret string."""
        secret = hmac.new(settings.SECRET_KEY.encode(), ('payid ' + str(pk)).encode()).hexdigest()
        return settings.PAYMENTS_REALM + ' ' + str(pk) + ' ' + secret

    @staticmethod
    def pk_from_custom(custom):
        """Restore the :class:`BaseTransaction` primary key from the secret "custom".

        Raises :class:`BaseTransaction.DoesNotExist` if the custom is wrong.

        Args:
            custom: A secret string.

        Returns:
            The primary key for :class:`BaseTransaction`."""
        r = custom.split(' ', 2)
        if len(r) != 3 or r[0] != settings.PAYMENTS_REALM:
            raise BaseTransaction.DoesNotExist
        try:
            pk = int(r[1])
            secret = hmac.new(settings.SECRET_KEY.encode(), ('payid ' + str(pk)).encode()).hexdigest()
            if r[2] != secret:
                raise BaseTransaction.DoesNotExist
            return pk
        except ValueError:
            raise BaseTransaction.DoesNotExist

    @abc.abstractmethod
    def invoice_id(self):
        """Invoice ID.

        Used internally to prevent more than one payment for the same transaction."""
        pass

    def invoiced_purchase(self):
        """Internal."""
        return self.purchase.old_subscription or self.purchase

    @abc.abstractmethod
    def subinvoice(self):
        """Subinvoice ID.

        Used internally to prevent more than one payment for the same transaction."""
        pass


class SimpleTransaction(BaseTransaction):
    """A one-time (non-recurring) transaction."""

    def subinvoice(self):
        return 1

    def invoice_id(self):
        return settings.PAYMENTS_REALM + ' p-%d' % (self.purchase.pk,)

    # Make transaction atomic to be sure that simpleitem.save() and advance_parent() do together
    @transaction.atomic
    def on_accept_regular_payment(self, email):
        """Handles confirmation of a (non-recurring) payment."""
        payment = SimplePayment.objects.create(transaction=self, email=email)
        self.purchase.status = SimplePaymentStatus.PAID
        self.purchase.payment = payment
        self.purchase.upgrade_subscription()
        self.purchase.save()
        try:
            self.advance_parent(self.purchase.simplepurchase.prolongpurchase, payment)
        except AttributeError:
            pass
        return payment


    @transaction.atomic
    def advance_parent(self, prolongpurchase, payment):
        """Advances the parent transaction on receive of a "prolong" payment.

        Args:
            prolongpurchase: :class:`ProlongPurchase`.

        `prolongitem.period` contains the number of days to advance the parent (:class:`SubscriptionItem`)
        item. The parent transaction is advanced this number of days.
        """
        parent_purchase = SubscriptionPurchase.objects.select_for_update().get(
            pk=prolongpurchase.prolonged_id)  # must be inside transaction
        # parent.email = transaction.email
        base_date = max(datetime.date.today(), parent_purchase.due_payment_date)
        klass = model_from_ref(payment.transaction.processor.klass)  # prolongpurchase.payment is None, so use payment instead
        parent_purchase.set_payment_date(klass.offset_date(base_date, prolongpurchase.period))
        parent_purchase.save()


class SubscriptionTransaction(BaseTransaction):
    """A transaction for a subscription service."""

    def subinvoice(self):
        return self.invoiced_purchase().subscriptionpurchase.subinvoice

    def invoice_id(self):
        if self.purchase.old_subscription:
            return settings.PAYMENTS_REALM + ' %d-%d-u' % (self.purchase.pk, self.subinvoice())
        else:
            return settings.PAYMENTS_REALM + ' %d-%d' % (self.purchase.pk, self.subinvoice())


class Item(models.Model):
    """Anything sold or rent.

    Apps using this package should create their product records manually.
    Then you create an instance of a subclass of this class before allowing
    the user to make a transaction.

    In a future we may provide an interface for registering new products.
    """

    product = models.ForeignKey('Product', null=True, on_delete=models.CASCADE)
    """The sold product."""

    product_qty = models.IntegerField(default=1)
    """Quantity of the sold product (often 1)."""

    currency = models.CharField(max_length=3, default='USD')
    """The currency for which this is sold."""

    price = models.DecimalField(max_digits=10, decimal_places=2)
    """Price of the item.
    
    For recurring payment it is the amount of one payment."""

    def __repr__(self):
        return "<Item pk=%d, %s>" % (self.pk, self.product.name)

    def __str__(self):
        return self.product.name

    @abc.abstractmethod
    def is_subscription(self):
        pass
    """Is this a recurring (or one-time) payment."""


class SimplePaymentStatus(object):
    NOT_PAID = 1
    PAID = 2
    REFUNDED = 3


class SimpleItem(Item):
    """Non-subscription item.

    To sell a non-subscription item, create a subclass of this model, describing your sold good."""

    def is_subscription(self):
        return False


class SubscriptionItem(Item):
    """Subscription (recurring) item.

    To sell a subscription item, create a subclass of this model, describing your sold service."""

    grace_period = Period(unit=Period.UNIT_DAYS, count=20)
    """How much :attr:`payment_deadline` is above :attr:`due_payment_data`."""

    payment_period = Period(unit=Period.UNIT_MONTHS, count=1)
    """How often to pay (for automatic recurring payments)."""

    trial_period = Period(unit=Period.UNIT_MONTHS, count=0)
    """Trial period.
    
    It may be zero."""

    def is_subscription(self):
        return True


class Purchase(models.Model):
    item = models.ForeignKey('Item', null=False, on_delete=models.CASCADE)

    parent = models.ForeignKey('AggregatePurchase', null=True, on_delete=models.SET_NULL, related_name='childs')
    """This purchase is a part of a composite purchase."""

    creation_date = models.DateTimeField(auto_now_add=True)
    """Date of item creation."""

    payment = models.OneToOneField('Payment', null=True, on_delete=models.CASCADE)
    """Payment accomplished for this item or `None`."""

    blocked = models.BooleanField(default=False)
    """A hacker or misbehavior detected."""

    gratis = models.BooleanField(default=False)
    """Provide a product or service for free."""

    shipping = models.DecimalField(max_digits=10, decimal_places=2, default=0)
    """Price of shipping.

    Remain zero if doubt."""

    tax = models.DecimalField(max_digits=10, decimal_places=2, default=0)
    """Tax for the sale.

    Remain zero if doubt."""

    # code = models.CharField(max_length=255) # TODO

    reminders_sent = models.SmallIntegerField(default=0, db_index=True)
    """Email (or SMS, etc.) payment reminders sent state.

    * 0 - no reminder sent
    * 1 - before due payment sent
    * 2 - at due payment sent
    * 3 - day before deadline sent

    TODO: Move to :class:`SubscriptionPurchase`?"""

    old_subscription = models.ForeignKey('Purchase', null=True, related_name='new_subscription',
                                         on_delete=models.CASCADE)
    """We remove old_subscription (if not `None`) automatically when new subscription is created.

    The new payment may be either one-time (:class:`SimpleItem` (usually :class:`ProlongPurchase`))
    or subscription (:class:`SubscriptionItem`)."""

    def __repr__(self):
        return "<Purchase pk=%d, %s>" % (self.pk, self.item.product.name)

    @property
    def is_aggregate(self):
        return False

    @transaction.atomic
    def upgrade_subscription(self):
        """Internal.

        It cancels the old subscription (if any).

        It can be called from both subscription IPN and payment IPN, so prepare to handle it two times."""
        if self.old_subscription:
            self.do_upgrade_subscription()

    def do_upgrade_subscription(self):
        """Internal.

        TODO: Remove ALL old subscriptions as in payment_system2."""
        try:
            self.old_subscription.subscriptionpurchase.force_cancel(is_upgrade=True)
        except CannotCancelSubscription:
            pass
        # self.on_upgrade_subscription(transaction, item.old_subscription)  # TODO: Needed?
        Purchase.objects.filter(pk=self.pk).update(old_subscription=None)

    # TODO: Move to Payment class?
    def send_rendered_email(self, template_name, subject, data):
        """Internal."""
        email = None
        try:
            email = self.payment.email
            # Item.objects.filter(pk=self.pk).update(email=email)
        except AttributeError:  # no .payment
            return
        if email is not None:
            html = render_to_string(template_name, data, request=None, using=None)
            text = html2text.html2text(html)
            send_mail(subject, text, settings.FROM_EMAIL, [email], html_message=html)


class SimplePurchase(Purchase):
    status = models.SmallIntegerField(_('Payment status'), default=SimplePaymentStatus.NOT_PAID)  # SimplePaymentStatus

    @property
    def paid(self):
        """It was paid by the user (and not refunded)."""
        cur = self
        while True:
            if cur._paid:
                return True
            cur = cur.parent.only('status', 'parent')
            if not cur:
                return False

    @property
    def _paid(self):
        """Internal."""
        return self.status == SimplePaymentStatus.PAID

    def is_paid(self):
        return (self.paid or self.gratis) and not self.blocked
    """If to consider the item paid (or gratis) but not blocked."""


class SubscriptionPurchase(Purchase):
    due_payment_date = models.DateField(default=datetime.date.today, db_index=True)
    """The reference payment date."""

    payment_deadline = models.DateField(null=True, db_index=True)  # may include "grace period"
    """The dealine payment date.

    After it is reached, the item is considered inactive."""

    trial = models.BooleanField(default=False, db_index=True)
    """Now in trial period."""

    # https://bitbucket.org/arcamens/django-payments/wiki/Invoice%20IDs
    subinvoice = models.PositiveIntegerField(default=1)  # no need for index, as it is used only at PayPal side
    """Internal."""

    subscription_reference = models.CharField(max_length=255, null=True)
    """As `recurring_payment_id` in PayPal.

    TODO: Avangate has it for every product, but PayPal for transaction as a whole."""

    processor = models.ForeignKey(PaymentProcessor, on_delete=models.CASCADE, null=True)
    """Payment processor for a subscription payment."""

    email = models.EmailField(null=True)
    """User's email.

    DalPay requires to notify the customer 10 days before every payment."""

    @property
    def subscribed(self):
        """Is in automatic (not manual) recurring mode."""
        return bool(self.subscription_reference)

    def is_active(self):
        """Is the item active (paid on time and not blocked).

        Usually you should use quick_is_active() instead because that is faster."""
        prior = self.payment_deadline is not None and \
                datetime.date.today() <= self.payment_deadline
        return (prior or self.gratis) and not self.blocked

    @staticmethod
    def quick_is_active(item_id):
        """Is the item with given PK active (paid on time and not blocked).

        Usually you should use quick_is_active() instead because that is faster."""
        item = SubscriptionItem.objects.filter(pk=item_id).\
            only('payment_deadline', 'gratis', 'blocked').get()
        return item.is_active()

    def set_payment_date(self, date):
        """Sets both :attr:`due_payment_date` and :attr:`payment_deadline`."""
        self.due_payment_date = date
        # klass = model_from_ref(self.payment.transaction.processor.klass)
        # self.payment_deadline = klass.offset_date(self.due_payment_date, self.grace_period)
        self.payment_deadline = self.due_payment_date + period_to_delta(self.item.subscriptionitem.grace_period)

    def start_trial(self):
        """Start trial period.

        This should be called after setting non-zero :attr:`trial_period`."""
        if self.item.subscriptionitem.trial_period.count != 0:
            self.trial = True
            # klass = model_from_ref(self.payment.transaction.processor.klass)  # not yet defined
            # self.set_payment_date(klass.offset_date(datetime.date.today(), self.trial_period))
            self.set_payment_date(datetime.date.today() + period_to_delta(self.item.subscriptionitem.trial_period))

    # TODO: The same as in do_upgrade_subscription()
    #@shared_task  # PayPal tormoz, so run in a separate thread # TODO: celery (with `TypeError: force_cancel() missing 1 required positional argument: 'self'`)
    def force_cancel(self, is_upgrade=False):
        """Cancels the :attr:`transaction`."""
        if self.subscription_reference:
            klass = model_from_ref(self.processor.klass)
            api = klass().api()
            try:
                api.cancel_agreement(self.subscription_reference, is_upgrade=is_upgrade)  # may raise an exception
            except CannotCancelSubscription:
                logger.warn("Cannot cancel subscription " + self.subscription_reference)
                # fallback
                SubscriptionPurchase.objects.filter(pk=self.pk).update(
                    payment=None, processor=None, subscription_reference=None, subinvoice=F('subinvoice') + 1)
                raise
            # transaction.cancel_subscription()  # runs in the callback
        else:
            # SubscriptionItem.objects.filter(payment=self.pk).update(payment=None, subinvoice=F('subinvoice') + 1)  # called in cancel_subscription()
            pass

    @django.db.transaction.atomic
    def activate_subscription(self, ref, email, processor):
        """Internal.

        "Competes" with :meth:`on_accept_regular_payment`."""
        SubscriptionPurchase.objects.filter(pk=self.pk).update(subscription_reference=ref, email=email, processor=processor)

    def cancel_subscription(self):
        """Called when we detect that the subscription was canceled."""
        # atomic operation
        SubscriptionPurchase.objects.filter(pk=self.pk).update(
            payment=None, subscription_reference=None, processor=None, subinvoice=F('subinvoice') + 1)
        if not self.old_subscription:  # don't send this email on plan upgrade
            self.cancel_subscription_email()

    def cancel_subscription_email(self):
        """Internal.

        Sends cancel subscription email."""
        url = settings.PAYMENTS_HOST + reverse(settings.PROLONG_PAYMENT_VIEW, args=[self.pk])
        days_before = (self.due_payment_date - datetime.date.today()).days
        self.send_rendered_email('debits/email/subscription-canceled.html',
                                 _("Service subscription canceled"),
                                 {'self': self,
                                  'product': self.item.product.name,
                                  'url': url,
                                  'days_before': days_before})

    @staticmethod
    def send_reminders():
        """Send all email reminders."""
        SubscriptionPurchase.send_regular_reminders()
        SubscriptionPurchase.send_trial_reminders()

    @staticmethod
    def send_regular_reminders():
        """Internal."""
        # start with the last
        SubscriptionPurchase.send_regular_before_due_reminders()
        SubscriptionPurchase.send_regular_due_reminders()
        SubscriptionPurchase.send_regular_deadline_reminders()

    @staticmethod
    def send_regular_before_due_reminders():
        """Internal."""
        days_before = settings.PAYMENTS_DAYS_BEFORE_DUE_REMIND
        reminder_date = datetime.date.today() + datetime.timedelta(days=days_before)
        q = SubscriptionPurchase.objects.filter(reminders_sent__lt=3, due_payment_date__lte=reminder_date, trial=False)
        for purchase in q:
            Item.objects.filter(pk=purchase.pk).update(reminders_sent=3)
            url = reverse(settings.PROLONG_PAYMENT_VIEW, args=[purchase.pk])
            purchase.send_rendered_email('debits/email/before-due-remind.html',
                                     _("You need to pay for %s") % purchase.product.name,
                                     {'transaction': purchase,
                                      'product': purchase.product.name,
                                      'url': url,
                                      'days_before': days_before})

    @staticmethod
    def send_regular_due_reminders():
        """Internal."""
        reminder_date = datetime.date.today()
        q = SubscriptionPurchase.objects.filter(reminders_sent__lt=2, due_payment_date__lte=reminder_date, trial=False)
        for purchase in q:
            Item.objects.filter(pk=purchase.pk).update(reminders_sent=2)
            url = reverse(settings.PROLONG_PAYMENT_VIEW, args=[purchase.pk])
            purchase.send_rendered_email('debits/email/due-remind.html',
                                     _("You need to pay for %s") % purchase.product.name,
                                     {'transaction': purchase,
                                      'product': purchase.product.name,
                                      'url': url})

    @staticmethod
    def send_regular_deadline_reminders():
        """Internal."""
        reminder_date = datetime.date.today()
        q = SubscriptionPurchase.objects.filter(reminders_sent__lt=1, payment_deadline__lte=reminder_date, trial=False)
        for purchase in q:
            Item.objects.filter(pk=purchase.pk).update(reminders_sent=1)
            url = reverse(settings.PROLONG_PAYMENT_VIEW, args=[purchase.pk])
            purchase.send_rendered_email('debits/email/deadline-remind.html',
                                     _("You need to pay for %s") % purchase.product.name,
                                     {'transaction': purchase,
                                      'product': purchase.product.name,
                                      'url': url})

    @staticmethod
    def send_trial_reminders():
        """Internal."""
        # start with the last
        SubscriptionPurchase.send_trial_before_due_reminders()
        SubscriptionPurchase.send_trial_due_reminders()
        SubscriptionPurchase.send_trial_deadline_reminders()

    @staticmethod
    def send_trial_before_due_reminders():
        """Internal."""
        days_before = settings.PAYMENTS_DAYS_BEFORE_TRIAL_END_REMIND
        reminder_date = datetime.date.today() + datetime.timedelta(days=days_before)
        q = SubscriptionPurchase.objects.filter(reminders_sent__lt=3, due_payment_date__lte=reminder_date, trial=True)
        for purchase in q:
            Item.objects.filter(pk=purchase.pk).update(reminders_sent=3)
            url = reverse(settings.PROLONG_PAYMENT_VIEW, args=[purchase.pk])
            purchase.send_rendered_email('debits/email/before-due-remind.html',
                                     _("You need to pay for %s") % purchase.product.name,
                                     {'transaction': purchase,
                                      'product': purchase.product.name,
                                      'url': url,
                                      'days_before': days_before})

    @staticmethod
    def send_trial_due_reminders():
        """Internal."""
        reminder_date = datetime.date.today()
        q = SubscriptionPurchase.objects.filter(reminders_sent__lt=2, due_payment_date__lte=reminder_date, trial=True)
        for purchase in q:
            Item.objects.filter(pk=purchase.pk).update(reminders_sent=2)
            url = reverse(settings.PROLONG_PAYMENT_VIEW, args=[purchase.pk])
            purchase.send_rendered_email('debits/email/due-remind.html',
                                     _("You need to pay for %s") % purchase.product.name,
                                     {'transaction': purchase,
                                      'product': purchase.product.name,
                                      'url': url})

    @staticmethod
    def send_trial_deadline_reminders():
        """Internal."""
        reminder_date = datetime.date.today()
        q = SubscriptionPurchase.objects.filter(reminders_sent__lt=1, payment_deadline__lte=reminder_date, trial=True)
        for purchase in q:
            Item.objects.filter(pk=purchase.pk).update(reminders_sent=1)
            url = reverse(settings.PROLONG_PAYMENT_VIEW, args=[purchase.pk])
            purchase.send_rendered_email('debits/email/deadline-remind.html',
                                     _("You need to pay for %s") % purchase.product.name,
                                     {'transaction': purchase,
                                      'product': purchase.product.name,
                                      'url': url})

    # TODO
    # def get_email(self):
    #     try:
    #         # We get the first email, as normally we have no more than one non-canceled transaction
    #         t = self.transactions.filter(subscription__canceled=False)[0]
    #         payment = AutomaticPayment.objects.filter(transaction=t).order_by('-id')[0]
    #         return payment.email
    #     except IndexError:  # no object
    #         return None


class ProlongPurchase(SimplePurchase):
    """Prolong :attr:`prolonged` item.

    This is meant to be a one-time payment which prolongs a manual subscription item."""

    prolonged = models.ForeignKey('SubscriptionPurchase', related_name='child', parent_link=False, on_delete=models.CASCADE)
    """Which subscription item to prolong."""

    period = Period(unit=Period.UNIT_MONTHS, count=0)
    """The amount of days (or weeks, months, etc.) how much to prolong."""

    def refund_payment(self):
        """Handle payment refund.

        For :class:`ProlongPurchase` we subtract the prolong days back from the :attr:`parent` item."""
        prolong2 = self.period
        prolong2.count *= -1
        klass = model_from_ref(self.payment.transaction.processor.klass)
        self.prolonged.set_payment_date(klass.offset_date(self.prolonged.due_payment_date, prolong2))
        self.prolonged.save()


class Payment(models.Model):
    """Base class describing a particular payment.

    It generated by our IPN handler."""

    payment_time = models.DateTimeField(_('Payment time'), auto_now_add=True)

    transaction = models.OneToOneField('BaseTransaction', on_delete=models.CASCADE)
    """The transaction we accepted."""

    email = models.EmailField(null=True)
    """User's email.

    DalPay requires to notify the customer 10 days before every payment."""

    def refund_payment(self):
        """Handles payment refund."""
        # Controversial decision to reset payment=None on refund
        # try:
        #     SimplePayment.objects.filter(pk=self.pk).update(payment=None, status=SimplePaymentStatus.REFUNDED)
        # except ObjectDoesNotExist:
        #     Payment.objects.filter(pk=self.pk).update(payment=None)
        try:
            SimplePurchase.objects.filter(pk=self.purchase.pk).update(status=SimplePaymentStatus.REFUNDED)
        except ObjectDoesNotExist:
            pass
        try:
            self.transaction.purchase.simplepurchase.prolongpurchase.refund_payment()
        except (SimplePurchase.DoesNotExist, ProlongPurchase.DoesNotExist):
            pass


class SimplePayment(Payment):
    """Non-recurring payment."""

    pass


class AutomaticPayment(Payment):
    """Automatic (recurring) payment."""

    processor = models.ForeignKey(PaymentProcessor, on_delete=models.CASCADE)
    """Payment processor."""

    subscription_reference = models.CharField(max_length=255, null=True)
    """As `recurring_payment_id` in PayPal.

    TODO: Avangate has it for every product, but PayPal for transaction as a whole."""

    # A transaction should have a code that identifies it.
    # code = models.CharField(max_length=255)


class AggregateItem(SimpleItem):
    """Several payments in one.

    TODO: Not tested!"""

    def calc(self):
        """Update price to be the sum of all children."""
        price = 0.0
        for child in self.childs.all():
            price += child.price
        self.price = price
        self.save()


class AggregatePurchase(SimplePurchase):
    """Several payments in one.

    TODO: Not tested!"""

    def calc(self):
        """Update shipping and tax to be the sum of all children.

        TODO: Also tax."""
        self.item.simpleitem.aggregateitem.calc()
        shipping = 0.0
        tax = 0.0
        for child in self.childs.all():
            shipping += child.shipping
            tax += child.tax
        self.shipping = shipping
        self.tax = tax
        self.save()

    @property
    def is_aggregate(self):
        return True


class CannotCancelSubscription(Exception):
    """Canceling subscription failed."""
    pass


class CannotRefund(Exception):
    """Refunding payment failed."""
    pass



import datetime

from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render, reverse
from django.utils.translation import ugettext_lazy as _
from .models import Organization, MyPurchase, PricingPlan
from .forms import CreateOrganizationForm, SwitchPricingPlanForm
from .business import create_organization
from debits.debits_base.base import Period, period_to_string
from debits.debits_base.models import SimpleTransaction, SubscriptionTransaction, ProlongPurchase, SubscriptionItem, \
    logger, \
    CannotCancelSubscription, ProlongPurchase, SimpleItem
import debits
from .processors import MyPayPalForm


def transaction_payment_view(request, transaction_id):
    """A view initiated from a transaction."""
    transaction = SubscriptionTransaction.objects.get(pk=int(transaction_id))
    purchase = transaction.purchase
    organization = purchase.organization
    return do_organization_payment_view(request, purchase, organization)


def organization_payment_view(request, organization_id):
    """A view initiated for an organization."""
    organization = Organization.objects.get(pk=int(organization_id))
    purchase = organization.purchase
    return do_organization_payment_view(request, purchase, organization)


def do_organization_payment_view(request, purchase, organization):
    """The common pars of views for :func:`transaction_payment_view` and :func:`organization_payment_view`."""
    plan_form = SwitchPricingPlanForm({'pricing_plan': purchase.plan.pk})
    pp = MyPayPalForm(request)
    return render(request, 'debits_test/organization-payment-view.html',
                  {'organization_id': organization.pk,
                   'organization': organization.name,
                   'item_id': purchase.pk,
                   'email': purchase.payment.email if purchase.payment else None,
                   'gratis': purchase.gratis,
                   'active': purchase.is_active(),
                   'blocked': purchase.blocked,
                   'manual_mode': not purchase.subscribed,
                   'processor_name': purchase.processor.name if purchase.processor else None,
                   # only for automatic recurring payment
                   'plan': purchase.plan.name,
                   'trial': purchase.trial,
                   'trial_period': period_to_string(purchase.item.subscriptionitem.trial_period),
                   'due_date': purchase.due_payment_date,
                   'deadline': purchase.payment_deadline,
                   'price': purchase.item.price,
                   'currency': purchase.item.currency,
                   'payment_period': period_to_string(purchase.item.subscriptionitem.payment_period),
                   'plan_form': plan_form,
                   'can_switch_to_recurring': pp.ready_for_subscription(purchase),
                   'subscription_allowed_date': pp.subscription_allowed_date(purchase),
                   'subscription_reference': purchase.subscription_reference,
                   'subinvoice': purchase.subinvoice})


def create_organization_view(request):
    """The view to create an example organization."""
    if request.method == 'POST':
        form = CreateOrganizationForm(request.POST)
        if form.is_valid():
            trial_months = 1 if 'use_trial' in request.POST else 0
            organization = create_organization(request.POST['name'], int(request.POST['pricing_plan']), trial_months)
            return HttpResponseRedirect(reverse('organization-prolong-payment', args=[organization.pk]))
    else:
        form = CreateOrganizationForm()

    return render(request, 'debits_test/create-organization.html', {'form': form})


def get_processor(request, hash):
    """Determine the payment processor, from a form."""
    processor_name = hash.pop('arcamens_processor')
    if processor_name == 'PayPal':
        form = MyPayPalForm(request)
        processor_id = debits.debits_base.processors.PAYMENT_PROCESSOR_PAYPAL
        processor = debits.debits_base.models.PaymentProcessor.objects.get(pk=processor_id)
    else:
        raise RuntimeError("Unsupported payment form.")
    return form, processor


def do_subscribe(hash, form, processor, purchase):
    """Start subscription to our subscription purchase."""
    transaction = SubscriptionTransaction.objects.create(processor=processor, purchase=purchase)
    return form.make_purchase_from_form(hash, transaction)


def do_prolong(hash, form, processor, purchase):
    """Start prolonging our subscription purchase."""
    periods = int(hash['periods'])
    subitem = SimpleItem.objects.create(product=purchase.item.product,
                                        currency=purchase.item.currency,
                                        price=purchase.item.price * periods)
    subpurchase = ProlongPurchase.objects.create(item=subitem,
                                                 prolonged=purchase,
                                                 period_unit=Period.UNIT_MONTHS,
                                                 period_count=periods)
    subtransaction = SimpleTransaction.objects.create(processor=processor, purchase=subpurchase)
    return form.make_purchase_from_form(hash, subtransaction)


def upgrade_calculate_new_period(k, purchase):
    """New period (in days) after an upgrade."""
    if purchase.due_payment_date:
        period = (purchase.due_payment_date - datetime.date.today()).days
    else:
        period = 0
    return round(period / k) if k > 1 else period  # don't increase paid period when downgrading


def upgrade_create_new_item(old_purchase, plan, new_period, organization):
    """Create new purchase used to upgrade another purchase (:obj:`old_purchase`)."""
    item = debits.debits_base.models.SubscriptionItem.objects.create(
        product=plan.product,
        currency=plan.currency,
        price=plan.price,
        payment_period_unit=Period.UNIT_MONTHS,
        payment_period_count=1,
        trial_period_unit=Period.UNIT_DAYS,
        trial_period_count=new_period)
    purchase = MyPurchase(item=item,
                          for_organization=organization,
                          plan=plan)
    purchase.set_payment_date(datetime.date.today() + datetime.timedelta(days=new_period))
    if old_purchase.subscribed:
        purchase.old_subscription = old_purchase
    purchase.save()
    return purchase


def do_upgrade(hash, form, processor, purchase, organization):
    """Start upgrading a subscription purchase,"""
    plan = PricingPlan.objects.get(pk=int(hash.pop('pricing_plan')))
    if plan.currency != purchase.item.currency:
        raise RuntimeError(_("Cannot upgrade to a payment plan with other currency."))
    if purchase.item.subscriptionitem.payment_period.unit != Period.UNIT_MONTHS or purchase.item.subscriptionitem.payment_period.count != 1:
        raise RuntimeError(_("Only one month payment period supported."))

    k = plan.price / purchase.item.price  # price multiplies
    new_period = upgrade_calculate_new_period(k, purchase)

    new_purchase = upgrade_create_new_item(purchase, plan, new_period, organization)

    if not purchase.subscribed:
        # Simply create a new purchase which can be paid later
        organization.purchase = new_purchase
        organization.save()
        return HttpResponseRedirect(reverse('organization-prolong-payment', args=[organization.pk]))
    else:
        upgrade_transaction = SubscriptionTransaction.objects.create(processor=processor, purchase=new_purchase)
        return form.make_purchase_from_form(hash, upgrade_transaction)


def purchase_view(request):
    """The main test view to make purchases, subscriptions, upgrades."""
    hash = request.POST.dict()
    op = hash.pop('arcamens_op')
    form, processor = get_processor(request, hash)
    organization_pk = int(hash.pop('organization'))  # in real code should use user login information
    organization = Organization.objects.get(pk=organization_pk)
    purchase = organization.purchase
    if op == 'subscribe':
        due_date = purchase.due_payment_date
        if due_date < datetime.date.today():
            due_date = datetime.date.today()
        new_item = debits.debits_base.models.SubscriptionItem.objects.create(
            product=purchase.plan.product,
            currency=purchase.plan.currency,
            price=purchase.plan.price,
            payment_period_unit=Period.UNIT_MONTHS,
            payment_period_count=1,
            trial_period_unit=Period.UNIT_DAYS,
            trial_period_count=(due_date - datetime.date.today()).days)
        new_purchase = MyPurchase(item=new_item,
                                  for_organization=organization,
                                  plan=purchase.plan)
        new_purchase.set_payment_date(due_date)
        new_purchase.save()
        return do_subscribe(hash, form, processor, new_purchase)
    elif op == 'manual':
        return do_prolong(hash, form, processor, purchase)
    elif op == 'upgrade':
        return do_upgrade(hash, form, processor, purchase, organization)


# TODO: purchase argument is not used
def do_unsubscribe(purchase):
    try:
        purchase.force_cancel()
    except CannotCancelSubscription as e:
        return HttpResponse(e)
    else:
        return HttpResponse('')  # empty string means success


def unsubscribe_organization_view(request, organization_pk):
    """Django view for the "Unsubscribe" button."""
    organization_pk = int(organization_pk)  # in real code should use user login information
    organization = Organization.objects.get(pk=organization_pk)
    purchase = organization.purchase.subscriptionpurchase
    return do_unsubscribe(purchase)
    # return HttpResponseRedirect(reverse('organization-prolong-payment', args=[organization.pk]))


def list_organizations_view(request):
    """Django view to list all the organizations."""
    list = [{'id': o.id, 'name': o.name} for o in Organization.objects.all()]
    return render(request, 'debits_test/list-organizations.html',
                  {'organizations': list})



import abc
import datetime
from django.urls import reverse
from debits.debits_base.processors import BasePaymentProcessor
from debits.debits_base.base import Period
from debits.debits_base.models import BaseTransaction
from django.conf import settings


# TODO:
# https://developer.paypal.com/docs/classic/express-checkout/integration-guide/ECRecurringPayments/
# You can increase the profile amount by only 20% in each 180-day interval after you create the profile.

class PayPalForm(BasePaymentProcessor):
    """Base class for processing submit of a PayPal form."""
    @classmethod
    def ipn_url(cls):
        return settings.IPN_HOST + reverse(cls.ipn_name())

    @abc.abstractclassmethod
    def ipn_name(cls):
        """Django view name for PayPal IPN."""
        pass

    def amend_hash_new_purchase(self, transaction, hash):
        # https://developer.paypal.com/docs/classic/paypal-payments-standard/integration-guide/Appx_websitestandard_htmlvariables/

        cart = hash.pop('arcamens_cart', transaction.purchase.is_aggregate)

        items = self.init_items(transaction)
        # if transaction.purchase.item.is_subscription():
        if hasattr(transaction, 'subscriptiontransaction'):
            self.make_subscription(items, transaction, transaction.purchase)
        else:
            self.make_regular(items, transaction, transaction.purchase, cart)

        items.update(hash)
        items['bn'] = 'Arcamens_SP_EC'  # we don't want this token be changed without changing the code
        return items

    def init_items(self, transaction):
        debug = settings.PAYPAL_DEBUG
        url = 'https://www.sandbox.paypal.com' if debug else 'https://www.paypal.com'
        return {'business': settings.PAYPAL_ID,
                'arcamens_action': url + "/cgi-bin/webscr",
                'cmd': "_xclick-subscriptions" if hasattr(transaction, 'subscriptiontransaction') else "_xclick",
                'notify_url': self.ipn_url(),
                'custom': BaseTransaction.custom_from_pk(transaction.pk),
                'invoice': transaction.invoice_id()}

    def make_subscription(self, items, transaction, purchase):
        """Internal."""
        items['item_name'] = self.product_name(purchase)
        items['src'] = 1

        unit_map = {Period.UNIT_DAYS: 'D',
                    Period.UNIT_WEEKS: 'W',
                    Period.UNIT_MONTHS: 'M',
                    Period.UNIT_YEARS: 'Y'}
        if purchase.item.subscriptionitem.trial_period.count > 0:
            items['a1'] = 0
            items['p1'] = purchase.item.subscriptionitem.trial_period.count
            items['t1'] = unit_map[purchase.item.subscriptionitem.trial_period.unit]
        items['a3'] = purchase.item.price + purchase.shipping + purchase.tax
        items['p3'] = purchase.item.subscriptionitem.payment_period.count
        items['t3'] = unit_map[purchase.item.subscriptionitem.payment_period.unit]

    def make_regular(self, items, transaction, purchase, cart):
        """Internal."""
        if cart:
            items['upload'] = 1
            i = 1
            for child in purchase.aggregatepurchase.childs.order_by('pk') if purchase.is_aggregate else [purchase]:
                items['item_name_' + str(i)] = self.product_name(child)
                items['amount_' + str(i)] = child.item.price
                items['shipping_' + str(i)] = child.shipping
                items['tax_' + str(i)] = child.tax
                items['quantity_' + str(i)] = child.item.product_qty
                i += 1
        else:
            items['item_name'] = self.product_name(purchase)[0:127]
            items['amount'] = purchase.item.price
            items['shipping'] = purchase.shipping
            items['tax'] = purchase.tax
            items['quantity'] = purchase.item.product_qty

    def subscription_allowed_date(self, purchase):
        return max(datetime.date.today(),
                   purchase.due_payment_date - datetime.timedelta(days=89))  # intentionally one day added to be sure

# TODO: Support Express Checkout
# https://developer.paypal.com/docs/classic/products/express-checkout/



import traceback
from decimal import Decimal
import datetime
import requests
from django.utils import timezone
from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from debits.debits_base.processors import PaymentCallback, PAYMENT_PROCESSOR_PAYPAL
from debits.debits_base.base import logger
from debits.debits_base.models import BaseTransaction, SimpleTransaction, SubscriptionTransaction, AutomaticPayment, \
    SubscriptionPurchase
from debits.debits_base.base import Period
from django.conf import settings


# https://www.angelleye.com/paypal-recurring-payments-reference-transactions-and-preapproved-payments/

# https://developer.paypal.com/docs/classic/ipn/integration-guide/IPNIntro/
# https://developer.paypal.com/docs/classic/ipn/integration-guide/IPNandPDTVariables/

# Examples of IPN for recurring payments:
# https://www.angelleye.com/paypal-recurring-payments-ipn-samples/
# https://gist.github.com/thenbrent/3037967

# How to test IPN:
# https://developer.paypal.com/docs/classic/ipn/integration-guide/IPNTesting/

# https://developer.paypal.com/docs/classic/products/recurring-payments/
# says that recurring payments are available for PayPal Payments Standard

# https://developer.paypal.com/docs/classic/express-checkout/integration-guide/ECRecurringPayments/
# for a general introduction into recurring payments in PayPal

# This is a trouble: https://developer.paypal.com/docs/classic/express-checkout/integration-guide/ECRecurringPayments/
# "For recurring payments with the Express Checkout API, PayPal does not allow certain updates, such as billing amount, within 3 days of the scheduled billing date."

# Recurring payments cannot be created for buyers in Germany or China. In this case, you can use reference transactions as an alternate solution:
# https://developer.paypal.com/docs/classic/express-checkout/integration-guide/ECReferenceTxns/


# # Internal
# def parse_date(string):
#     # Not tread safe!!
#     # old_locale = locale.getlocale(locale.LC_TIME)
#     # locale.setlocale(locale.LC_TIME, "C")
#     ret = datetime.datetime.strptime(string, '%H:%M:%S %b %d, %Y %Z')
#     # locale.setlocale(locale.LC_TIME, old_locale)
#     return ret


# Internal.
from debits.paypal.models import PayPalAPI, PayPalProcessorInfo

MONTHS = [
    'Jan', 'Feb', 'Mar', 'Apr',
    'May', 'Jun', 'Jul', 'Aug',
    'Sep', 'Oct', 'Nov', 'Dec',
]


# Based on https://github.com/spookylukey/django-paypal/blob/master/paypal/standard/forms.py
def parse_date(value):
    """Internal."""
    value = value.strip()  # needed?

    time_part, month_part, day_part, year_part, zone_part = value.split()
    month_part = month_part.strip(".")
    day_part = day_part.strip(",")
    month = MONTHS.index(month_part) + 1
    day = int(day_part)
    year = int(year_part)
    hour, minute, second = map(int, time_part.split(":"))
    dt = datetime(year, month, day, hour, minute, second)

    if zone_part in ["PDT", "PST"]:
        # PST/PDT is 'US/Pacific'
        dt = timezone.pytz.timezone('US/Pacific').localize(
            dt, is_dst=zone_part == 'PDT')
        if not settings.USE_TZ:
            dt = timezone.make_naive(dt, timezone=timezone.utc)
    return dt


# FIXME: Refund fails for coupon or gift certificates, because they support only full refunds

@method_decorator(csrf_exempt, name='dispatch')
class PayPalIPN(PaymentCallback, View):
    """This class processes all kinds of PayPal IPNs.

    All its methods are considered internal."""

    # See https://developer.paypal.com/docs/classic/express-checkout/integration-guide/ECRecurringPayments/
    # for all kinds of IPN for recurring payments.
    def post(self, request):
        try:
            self.do_post(request)
        except KeyError as e:
            logger.warning("PayPal IPN var %s is missing" % e)
        except:
            import traceback
            traceback.print_exc()
        return HttpResponse('', content_type="text/plain")

    def do_post(self, request):
        # 'payment_date', 'time_created' unused
        if request.POST['receiver_email'] == settings.PAYPAL_EMAIL:
            self.do_do_post(request.POST, request)
        else:
            logger.warning("Wrong PayPal email")

    def do_do_post(self, POST, request):
        debug = settings.PAYPAL_DEBUG
        url = 'https://www.sandbox.paypal.com' if debug else 'https://www.paypal.com'
        r = requests.post(url + '/cgi-bin/webscr',
                          'cmd=_notify-validate&' + request.body.decode(
                              POST.get('charset') or request.content_params['charset']),
                          headers={
                              'content-type': request.content_type})  # message must use the same encoding as the original
        if r.text == 'VERIFIED':
            self.verified_post(POST, request)
        else:
            logger.warning("PayPal verification not passed")

    def verified_post(self, POST, request):
        # print('custom', POST['custom'])  # Don't print sensitive data
        transaction_id = BaseTransaction.pk_from_custom(POST['custom'])
        self.on_transaction_complete(POST, transaction_id)

    def on_transaction_complete(self, POST, transaction_id):
        # Crazy: Recurring payment and subscription debits are not the same.
        # 'recurring_payment_id' and 'subscr_id' are equivalent: https://thereforei.am/2012/07/03/cancelling-subscriptions-created-with-paypal-standard-via-the-express-checkout-api/
        type_dispatch = {
            'web_accept': self.accept_regular_payment,
            'cart': self.accept_regular_payment,
            'express_checkout': self.accept_regular_payment,
            'recurring_payment': self.accept_recurring_payment,
            'subscr_payment': self.accept_subscription_payment,
            'recurring_payment_profile_created': self.accept_recurring_signup,
            'subscr_signup': self.accept_subscription_signup,
            'recurring_payment_profile_cancel': self.accept_recurring_canceled,
            'recurring_payment_suspended': self.accept_recurring_canceled,
            'subscr_cancel': self.accept_recurring_canceled
        }
        if 'payment_status' in POST and POST['payment_status'] == 'Refunded':
            self.accept_refund(POST, transaction_id)
        else:
            type_dispatch[POST['txn_type']](POST, transaction_id)

    def accept_refund(self, POST, transaction_id):
        self.do_appect_refund(POST, transaction_id)

    def do_appect_refund(self, POST, transaction_id):
        try:
            transaction = BaseTransaction.objects.get(pk=transaction_id)
        except BaseTransaction.DoesNotExist:
            traceback.print_exc()
            return
        if POST['mc_currency'] == transaction.purchase.item.currency:
            transaction.payment.refund_payment()
        else:
            logger.warning("Wrong refund currency.")

    def accept_regular_payment(self, POST, transaction_id):
        if POST['payment_status'] == 'Completed':
            self.do_accept_regular_payment(POST, transaction_id)

    def do_accept_regular_payment(self, POST, transaction_id):
        POST = POST.dict()  # for POST.get() below
        self.do_do_accept_regular_payment(POST, transaction_id)

    def do_do_accept_regular_payment(self, POST, transaction_id):
        try:
            transaction = SimpleTransaction.objects.get(pk=transaction_id)
        except BaseTransaction.DoesNotExist:
            traceback.print_exc()
            return
        if Decimal(POST['mc_gross']) == transaction.purchase.item.price and \
                        Decimal(POST['shipping']) == transaction.purchase.shipping and \
                        Decimal(POST['tax']) == transaction.purchase.tax and \
                        POST['mc_currency'] == transaction.purchase.item.currency:
            if self.auto_refund(transaction, transaction.purchase.simplepurchase.prolongpurchase.prolonged, POST):
                return HttpResponse('')
            payment = transaction.on_accept_regular_payment(POST['payer_email'])
            self.on_payment(payment)
        else:
            logger.warning("Wrong amount or currency")

    def accept_recurring_payment(self, POST, transaction_id):
        if POST['payment_status'] != 'Completed':
            return
        self.do_accept_recurring_payment(POST, transaction_id)

    def do_accept_recurring_payment(self, POST, transaction_id):
        # transaction = BaseTransaction.objects.select_for_update().get(pk=transaction_id)  # only inside transaction
        try:
            transaction = SubscriptionTransaction.objects.get(pk=transaction_id)
        except BaseTransaction.DoesNotExist:
            traceback.print_exc()
            return
        if Decimal(POST['amount_per_cycle']) == transaction.purchase.item.price + transaction.purchase.item.shipping + transaction.purchase.item.tax and \
                        POST['payment_cycle'] in self.pp_payment_cycles(transaction.purchase.item):
            self.do_do_accept_subscription_or_recurring_payment(transaction, transaction.purchase.item, POST, POST['recurring_payment_id'])
        else:
            logger.warning("Wrong recurring payment data")

    def accept_subscription_payment(self, POST, transaction_id):
        if POST['payment_status'] != 'Completed':
            return
        self.do_accept_subscription_payment(POST, transaction_id)

    def do_do_accept_subscription_or_recurring_payment(self, transaction, purchase, POST, ref):
        if self.auto_refund(transaction, purchase, POST):
            return HttpResponse('')
        purchase.subscriptionpurchase.activate_subscription(ref, POST['payer_email'], PAYMENT_PROCESSOR_PAYPAL)
        # This is already done in activate_subscription():
        payment = AutomaticPayment.objects.create(transaction=transaction,
                                                  email=POST['payer_email'],
                                                  subscription_reference=ref,
                                                  processor_id=PAYMENT_PROCESSOR_PAYPAL)
        purchase.payment = payment
        self.do_subscription_or_recurring_payment(purchase.subscriptionpurchase)  # calls save()
        self.on_payment(transaction.payment.automaticpayment)

    def do_accept_subscription_payment(self, POST, transaction_id):
        # transaction = BaseTransaction.objects.select_for_update().get(pk=transaction_id)  # only inside transaction
        try:
            transaction = SubscriptionTransaction.objects.get(pk=transaction_id)
        except BaseTransaction.DoesNotExist:
            traceback.print_exc()
            return
        purchase = transaction.purchase
        if Decimal(POST['mc_gross']) == purchase.item.price + purchase.shipping + purchase.tax and \
                        POST['mc_currency'] == purchase.item.currency:
            self.do_do_accept_subscription_or_recurring_payment(transaction, purchase, POST, POST['subscr_id'])
        else:
            logger.warning("Wrong subscription payment data")

    def do_subscription_or_recurring_payment(self, purchase):
        # transaction.processor = PaymentProcessor.objects.get(pk=PAYMENT_PROCESSOR_PAYPAL)
        purchase.trial = False
        date = purchase.due_payment_date
        if purchase.item.subscriptionitem.payment_period.count > 0:  # hack to eliminate infinite loop
            while date <= datetime.date.today():
                date = self.advance_item_date(date, purchase)
        purchase.due_payment_date = date
        purchase.save()

    def advance_item_date(self, date, purchase):
        date = PayPalProcessorInfo.offset_date(date, purchase.item.subscriptionitem.payment_period)
        purchase.set_payment_date(date)
        purchase.reminders_sent = 0
        return date

    def do_subscription_or_recurring_created(self, transaction, POST, ref):
        purchase = transaction.purchase.subscriptionpurchase
        purchase.activate_subscription(ref, POST['payer_email'], PAYMENT_PROCESSOR_PAYPAL)
        # transaction.processor = PaymentProcessor.objects.get(pk=PAYMENT_PROCESSOR_PAYPAL)
        SubscriptionPurchase.objects.filter(pk=purchase.pk).update(trial=False)
        purchase.upgrade_subscription()
        self.on_subscription_created(POST, purchase)

    def accept_subscription_signup(self, POST, transaction_id):
        self.do_accept_subscription_signup(POST, transaction_id)

    def do_accept_subscription_signup(self, POST, transaction_id):
        try:
            transaction = SubscriptionTransaction.objects.get(pk=transaction_id)
        except BaseTransaction.DoesNotExist:
            traceback.print_exc()
            return
        purchase = transaction.purchase.subscriptionpurchase
        m = {
            Period.UNIT_DAYS: 'D',
            Period.UNIT_WEEKS: 'W',
            Period.UNIT_MONTHS: 'M',
            Period.UNIT_YEARS: 'Y',
        }
        period1_right = (purchase.item.subscriptionitem.trial_period.count == 0 and 'period1' not in POST) or \
                        (purchase.item.subscriptionitem.trial_period.count != 0 and 'period1' in POST and \
                         POST['period1'] == str(purchase.item.subscriptionitem.trial_period.count)+' '+m[purchase.item.subscriptionitem.trial_period.unit])
        if period1_right and 'period2' not in POST and \
                        Decimal(POST['amount3']) == purchase.item.price and \
                        POST['period3'] == str(purchase.item.subscriptionitem.payment_period.count)+' '+m[purchase.item.subscriptionitem.payment_period.unit] and \
                        POST['mc_currency'] == purchase.item.currency:
            self.do_subscription_or_recurring_created(transaction, POST, POST['subscr_id'])
        else:
            logger.warning("Wrong subscription signup data")

    def accept_recurring_signup(self, POST, transaction_id):
        try:
            transaction = SubscriptionTransaction.objects.get(pk=transaction_id)
        except BaseTransaction.DoesNotExist:
            traceback.print_exc()
            return
        if 'period1' not in POST and 'period2' not in POST and \
                        Decimal(POST['mc_amount3']) == transaction.purchase.item.price + transaction.purchase.shipping + transaction.purchase.tax and \
                        POST['mc_currency'] == transaction.purchase.item.currency and \
                        POST['period3'] in self.pp_payment_cycles(transaction):
            self.do_subscription_or_recurring_created(transaction, POST, POST['recurring_payment_id'])
        else:
            logger.warning("Wrong recurring signup data")

    def accept_recurring_canceled(self, POST, transaction_id):
        self.do_accept_recurring_canceled(POST, transaction_id)

    def do_accept_recurring_canceled(self, POST, transaction_id):
        try:
            transaction = SubscriptionTransaction.objects.get(pk=transaction_id)
        except BaseTransaction.DoesNotExist:
            traceback.print_exc()

            return
        transaction.purchase.subscriptionpurchase.cancel_subscription()
        self.on_subscription_canceled(POST, transaction.purchase)

    def auto_refund(self, transaction, purchase, POST):
        # "purchase" is SubscriptionItem
        if self.should_auto_refund():
            api = PayPalAPI()
            # FIXME: Wrong for American Express card: https://www.paypal.com/us/selfhelp/article/How-do-I-issue-a-full-or-partial-refund-FAQ780
            amount = (transaction.purchase.item.price - Decimal(0.30)).quantize(Decimal('1.00'))
            api.refund(POST['txn_id'], str(amount))
            return True
        return False

    def should_auto_refund(self):
        return False

    # Ugh, PayPal
    def pp_payment_cycles(self, purchase):
        first_tmpl = {
            Period.UNIT_DAYS: 'every %d Days',
            Period.UNIT_WEEKS: 'every %d Weeks',
            Period.UNIT_MONTHS: 'every %d Months',
            Period.UNIT_YEARS: 'every %d Years',
        }[purchase.item.subscriptionitem.payment_period.unit]
        first = first_tmpl % purchase.item.subscriptionitem.payment_period.count
        if purchase.item.subscriptionitem.payment_period.count == 1:
            second = {
                Period.UNIT_DAYS: 'Daily',
                Period.UNIT_WEEKS: 'Weekly',
                Period.UNIT_MONTHS: 'Monthly',
                Period.UNIT_YEARS: 'Yearly',
            }[purchase.item.subscriptionitem.payment_period.unit]
            return (first, second)
        else:
            return (first,)



from debits.debits_base.models import AutomaticPayment
from debits.paypal.views import PayPalIPN


class MyPayPalIPN(PayPalIPN):
    """Mixin to handle purchase events.

    Two subscription IPNs may call both :meth:`on_subscription_created` and :meth:`on_payment`.
    It is not a problem (if not to count a tiny performance lag).

    TODO: Generalize it for non PayPal processors."""
    def on_subscription_created(self, POST, purchase):
        self.do_purchase(purchase)

    def on_payment(self, payment):
        if isinstance(payment, AutomaticPayment):
            purchase = payment.transaction.purchase
            self.do_purchase(purchase)

    def do_purchase(self, purchase):
        """Set the :class:`~debits.debits_test.models.MyPurchase` for an :class:`~debits.debits_test.models.Organization`."""
        organization = purchase.subscriptionpurchase.mypurchase.for_organization
        if organization is not None:
            organization.purchase = purchase.subscriptionpurchase.mypurchase
            organization.save()



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2017-05-05 20:31
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('debits_base', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Organization',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='PricingPlan',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('price', models.DecimalField(decimal_places=2, max_digits=10)),
                ('currency', models.CharField(max_length=3)),
                ('period_unit', models.SmallIntegerField()),
                ('period_count', models.SmallIntegerField()),
                ('product', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='debits_base.Product')),
            ],
        ),
        migrations.CreateModel(
            name='MyPurchase',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('for_organization', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='for_purchase', to='debits_test.Organization')),
                ('item', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='debits_base.SubscriptionItem')),
                ('plan', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='debits_test.PricingPlan')),
            ],
        ),
        migrations.AddField(
            model_name='organization',
            name='purchase',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='debits_test.MyPurchase'),
        ),
    ]



import json

import requests
from dateutil.relativedelta import relativedelta

from debits.debits_base.base import Period, period_to_delta

try:
    from html import escape  # python 3.x
except ImportError:
    from cgi import escape  # python 2.x
from django.db import models
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from debits.debits_base.models import logger, CannotCancelSubscription, CannotRefund


class PayPalProcessorInfo(models.Model):
    class Meta:
        """Don't save it in the DB (use only for get_model())"""
        managed = False

    def api(self):
        return PayPalAPI()

    @staticmethod
    def offset_date(date, offset):
        """Used to calculate the next recurring payment date."""
        delta = period_to_delta(offset)
        new_date = date + delta
        if offset.unit in (Period.UNIT_MONTHS, Period.UNIT_YEARS) and new_date.day != date.day:
            new_date += relativedelta(days=1)
        return new_date


class PayPalAPI(object):
    """PayPal API.

    This code only provides a subset of the possible functionality, for
    something more comprehensive see https://github.com/paypal/PayPal-Python-SDK
    To login into PayPal we use a Bearer from https://api.paypal.com/v1/oauth2/token
    with secret from https://developer.paypal.com/developer/applications"""

    def __init__(self):
        """Creates a HTTP session to access PayPal API."""
        debug = settings.PAYPAL_DEBUG
        self.server = 'https://api.sandbox.paypal.com' if debug else 'https://api.paypal.com'
        s = requests.Session()
        s.headers.update({'Accept': 'application/json', 'Accept-Language': 'en_US'})
        r = s.post(self.server + '/v1/oauth2/token',
                   data='grant_type=client_credentials',
                   headers={'content-type': 'application/x-www-form-urlencoded'},
                   auth=(settings.PAYPAL_CLIENT_ID, settings.PAYPAL_SECRET))
        token = r.json()["access_token"]
        s.headers.update({'Authorization': 'Bearer '+token})
        self.session = s

    def cancel_agreement(self, agreement_id, is_upgrade=False):
        """Cancels a PayPal recurring payment."""
        note = _("Upgrading billing plan") if is_upgrade else _("Canceling a service")
        # https://developer.paypal.com/docs/api/#agreement_cancel
        # https://developer.paypal.com/docs/api/payments.billing-agreements#agreement_cancel
        logger.debug("PayPal: now canceling agreement %s" % escape(agreement_id))
        r = self.session.post(self.server + ('/v1/payments/billing-agreements/%s/cancel' % escape(agreement_id)),
                              data='{"note": "%s"}' % note,
                              headers={'content-type': 'application/json'})
        if r.status_code < 200 or r.status_code >= 300:  # PayPal returns 204, to be sure
            # Don't include secret information into the message
            raise CannotCancelSubscription(r.json()["message"])
            # raise RuntimeError(_("Cannot cancel a billing agreement at PayPal. Please contact support:\n" + r.json()["message"]))

    def refund(self, transaction_id, sum=None, currency='USD'):
        """Refunds a PayPal payment."""
        logger.debug("PayPal: now refunding transaction %s" % escape(transaction_id))
        data = {}
        if sum is not None:
            data['amount'] = {'total': sum, 'currency': currency}
        r = self.session.post(self.server + ('/v1/payments/sale/%s/refund' % escape(transaction_id)),
                              data=json.dumps(data),
                              headers = {'content-type': 'application/json'})
        if r.status_code < 200 or r.status_code >= 300:  # PayPal returns 204, to be sure
            # Don't include secret information into the message
            raise CannotRefund(r.json()["message"])
            # raise RuntimeError(_("Cannot cancel a billing agreement at PayPal. Please contact support:\n" + r.json()["message"]))


    # It does not work with PayPal subscriptions: https://www.paypal-knowledge.com/infocenter/index?page=content&id=FAQ1987&actp=LIST
    # def agreement_is_active(self, agreement_id):
    #     r = self.session.get(self.server + ('/v1/payments/billing-agreements/%s' % escape(agreement_id)),
    #                          headers={'content-type': 'application/json'})
    #     # ...



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2017-05-05 20:31
from __future__ import unicode_literals

import datetime
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='BaseTransaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_date', models.DateField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='Item',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_date', models.DateField(auto_now_add=True)),
                ('product_qty', models.IntegerField(default=1)),
                ('blocked', models.BooleanField(default=False)),
                ('currency', models.CharField(default='USD', max_length=3)),
                ('price', models.DecimalField(decimal_places=2, max_digits=10)),
                ('shipping', models.DecimalField(decimal_places=2, default=0, max_digits=10)),
                ('gratis', models.BooleanField(default=False)),
                ('reminders_sent', models.SmallIntegerField(db_index=True, default=0)),
            ],
        ),
        migrations.CreateModel(
            name='Payment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email', models.EmailField(max_length=254, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='PaymentProcessor',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('url', models.URLField(max_length=255)),
                ('api_app_label', models.CharField(max_length=100)),
                ('api_model', models.CharField(max_length=100, verbose_name='python model class name')),
            ],
        ),
        migrations.CreateModel(
            name='Product',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('subscription_reference', models.CharField(max_length=255, null=True)),
                ('email', models.EmailField(max_length=254, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='AutomaticPayment',
            fields=[
                ('payment_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='debits_base.Payment')),
            ],
            bases=('debits_base.payment',),
        ),
        migrations.CreateModel(
            name='SimpleItem',
            fields=[
                ('item_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='debits_base.Item')),
                ('paid', models.BooleanField(default=False)),
            ],
            bases=('debits_base.item',),
        ),
        migrations.CreateModel(
            name='SimplePayment',
            fields=[
                ('payment_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='debits_base.Payment')),
            ],
            bases=('debits_base.payment',),
        ),
        migrations.CreateModel(
            name='SimpleTransaction',
            fields=[
                ('basetransaction_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='debits_base.BaseTransaction')),
            ],
            bases=('debits_base.basetransaction',),
        ),
        migrations.CreateModel(
            name='SubscriptionItem',
            fields=[
                ('item', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='subscriptionitem', serialize=False, to='debits_base.Item')),
                ('due_payment_date', models.DateField(db_index=True, default=datetime.date.today)),
                ('payment_deadline', models.DateField(db_index=True, null=True)),
                ('last_payment', models.DateField(db_index=True, null=True)),
                ('trial', models.BooleanField(db_index=True, default=False)),
                ('grace_period_unit', models.SmallIntegerField(default=1)),
                ('grace_period_count', models.SmallIntegerField(default=20)),
                ('payment_period_unit', models.SmallIntegerField(default=3)),
                ('payment_period_count', models.SmallIntegerField(default=1)),
                ('trial_period_unit', models.SmallIntegerField(default=3)),
                ('trial_period_count', models.SmallIntegerField(default=0)),
                ('subinvoice', models.PositiveIntegerField(default=1)),
                ('active_subscription', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to='debits_base.Subscription')),
            ],
            bases=('debits_base.item',),
        ),
        migrations.CreateModel(
            name='SubscriptionTransaction',
            fields=[
                ('basetransaction_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='debits_base.BaseTransaction')),
                ('item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transactions', to='debits_base.SubscriptionItem')),
            ],
            bases=('debits_base.basetransaction',),
        ),
        migrations.AddField(
            model_name='item',
            name='old_subscription',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='new_subscription', to='debits_base.Subscription'),
        ),
        migrations.AddField(
            model_name='item',
            name='product',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='debits_base.Product'),
        ),
        migrations.AddField(
            model_name='basetransaction',
            name='processor',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='debits_base.PaymentProcessor'),
        ),
        migrations.CreateModel(
            name='ProlongItem',
            fields=[
                ('simpleitem_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='debits_base.SimpleItem')),
                ('prolong_unit', models.SmallIntegerField(default=3)),
                ('prolong_count', models.SmallIntegerField(default=0)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='child', to='debits_base.SubscriptionItem')),
            ],
            bases=('debits_base.simpleitem',),
        ),
        migrations.AddField(
            model_name='subscription',
            name='transaction',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='debits_base.SubscriptionTransaction'),
        ),
        migrations.AddField(
            model_name='simpletransaction',
            name='item',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transactions', to='debits_base.SimpleItem'),
        ),
        migrations.AddField(
            model_name='simplepayment',
            name='transaction',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='debits_base.SimpleTransaction'),
        ),
        migrations.AddField(
            model_name='automaticpayment',
            name='transaction',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='debits_base.SubscriptionTransaction'),
        ),
    ]



from django.shortcuts import render
from debits.paypal.form import PayPalForm
from debits.debits_base.models import ProlongPurchase

class MyPayPalForm(PayPalForm):
    """A mixin result."""

    def __init__(self, request):
        self.request = request

    @classmethod
    def ipn_name(cls):
        return 'paypal-ipn'

    def product_name(self, purchase):
        """What "product" PayPal shows for the purchase."""
        if isinstance(purchase, ProlongPurchase):
            purchase = purchase.prolonged
        return purchase.item.product.name + ': ' + purchase.mypurchase.plan.name



from django.conf.urls import url
from .callbacks import MyPayPalIPN
from . import views

urlpatterns = [
    url(r'^$', views.list_organizations_view, name='list-organizations'),
    url(r'^pay$', views.purchase_view, name='do-purchase'),
    url(r'^create-organization$', views.create_organization_view, name='create-organization'),
    url(r'^transaction-prolong-payment/([0-9]+)$', views.transaction_payment_view, name='transaction-prolong-payment'),
    url(r'^organization-prolong-payment/([0-9]+)$', views.organization_payment_view, name='organization-prolong-payment'),
    url(r'^unsubscribe-organization/([0-9]+)$', views.unsubscribe_organization_view, name='unsubscribe-organization'),
    url(r'^paypal/ipn$', MyPayPalIPN.as_view(), name='paypal-ipn')
]



from django.db import transaction

import debits
from debits.debits_base.base import Period
from .models import Organization, MyPurchase, PricingPlan


# TODO: Move .plan to Plan?
@transaction.atomic
def create_organization(name, pricing_plan_id, trial_months):
    """Creates a new example organization.

    It also associates a :class:`~debits.debits_test.models.MyPurchase` with it."""
    plan = PricingPlan.objects.get(pk=pricing_plan_id)
    item = debits.debits_base.models.SubscriptionItem.objects.create(product=plan.product,
                                                                     currency=plan.currency,
                                                                     price=plan.price,
                                                                     payment_period_unit=Period.UNIT_MONTHS,
                                                                     payment_period_count=1,
                                                                     trial_period_unit=Period.UNIT_MONTHS,
                                                                     trial_period_count=trial_months)
    purchase = MyPurchase(item=item, plan=plan)
    if trial_months:
        purchase.start_trial()
    purchase.save()
    org = Organization.objects.create(name=name, purchase=purchase)
    purchase.for_organization = org
    purchase.save()
    return org



"""A transaction_utils package.
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
# with open(path.join(here, 'README.md'), encoding='utf-8') as f:
#     long_description = f.read()

name = 'django-debits'
version = '0.1.7'
release = version

setup(
    name=name,

    version=version,

    description='Accepting payments',
    long_description='Accepting payments (currently supports PayPal)',

    # The project's main homepage.
    url='http://freesoft.portonvictor.org/django-debits.xml',

    # Author details
    author='Victor Porton',
    author_email='porton@narod.ru',

    # Choose your license
    license='AGPLv3+',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # What does your project relate to?
    keywords='payment paypal',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['test*']),

    command_options={
        'build_sphinx': {
            'project': ('setup.py', name),
            'version': ('setup.py', version),
            'release': ('setup.py', release),
            'source_dir': ('setup.py', 'doc')}},
)



import logging
from composite_field import CompositeField
from dateutil.relativedelta import relativedelta
from django.db import models
from django.utils.translation import ugettext_lazy as _


logger = logging.getLogger('debits')
"""The logger used by Debits."""


class Period(CompositeField):
    """Period (for example of recurring payment or of a trial subscription).

    It may be expressed in days, weeks, months, or years. Several units cannot be mixed:
    for example, it cannot be 2 months and 13 days."""
    UNIT_DAYS = 1
    UNIT_WEEKS = 2
    UNIT_MONTHS = 3
    UNIT_YEARS = 4

    period_choices = ((UNIT_DAYS, _("days")),  # different processors may support a part of it
                      (UNIT_WEEKS, _("weeks")),
                      (UNIT_MONTHS, _("months")),
                      (UNIT_YEARS, _("years")))
    """For Django :class:`~django.forms.ChoiceField`."""

    unit = models.SmallIntegerField()
    """days, weeks, months, or years."""

    count = models.SmallIntegerField()
    """The number of the units"""

    def __init__(self, unit=None, count=None):
        super().__init__()
        if unit is not None:
            self['unit'].default = unit
        if count is not None:
            self['count'].default = count


# The following functions do not work as a method, because
# CompositeField is replaced with composite_field.base.CompositeField.Proxy:

def period_to_string(period):
    """Human readable description of a period.

    Args:
        period: `Period` field.

    Returns:
        A human readable string."""
    hash = {e[0]: e[1] for e in Period.period_choices}
    return "%d %s" % (period.count, hash[period.unit])


def period_to_delta(period):
    """Convert :class:`Period` to :class:`relativedelta`."""
    return {
        Period.UNIT_DAYS: lambda: relativedelta(days=period.count),
        Period.UNIT_WEEKS: lambda: relativedelta(weeks=period.count),
        Period.UNIT_MONTHS: lambda: relativedelta(months=period.count),
        Period.UNIT_YEARS: lambda: relativedelta(years=period.count),
    }[period.unit]()



from django import forms
from django.utils.translation import ugettext_lazy as _
from .models import PricingPlan
from .products import PRODUCT_ITEM_1

class CreateOrganizationForm(forms.Form):
    """Django form to create a new example organization."""
    name = forms.CharField(label=_('Organization name'))
    pricing_plan = forms.ModelChoiceField(PricingPlan.objects.filter(product=PRODUCT_ITEM_1),
                                          label=_('Pricing plan'))
    use_trial = forms.BooleanField(label=_('With trial period'), required=False)


class SwitchPricingPlanForm(forms.Form):
    """Django form to switch pricing plan."""
    pricing_plan = forms.ModelChoiceField(PricingPlan.objects.filter(product=PRODUCT_ITEM_1))
